/*
  Performance Engineering Laboratory © 2017-2018
*/

unsigned long maxsize = 128 * 1024 * 1024; // 512 * ...works slowly
// 1 << 30 = 1073741824 there is space, but things become very slow

#if defined(MEASURE_TURNS)

long long turns = 0;

#endif

#if defined(MEASURE_COMPS) or defined(MEASURE_MOVES)

long long comps = 0;
long long moves = 0;

#endif

#include <cassert> // assert macro
#include <cstdlib> // std::atoi std::size_t
#include <cmath> // std::sqrt
#include <ctime> // std::clock_t std::clock CLOCKS_PER_SEC
#include <iostream> // std::cout std::cerr
#include <iterator> // std::iterator_traits std::distance
#include <limits> // std::numeric_limits
#include <random> // std::linear_congruential_engine

#include "algorithm.h++" // NAME::solve
#include "point.h++" // point

using linear_congruential_engine = std::linear_congruential_engine<
  unsigned long long, 6364136223846793005U, 1442695040888963407U, 0U>;
constexpr unsigned long long seed = 10164167376618180197U;
linear_congruential_engine random_number_generator{seed};

#if defined(DISC)

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  T t = std::numeric_limits<T>::max();
  std::uniform_int_distribution<T> distribution(-t, t);
  for (I q = p; q != r; ++q) {
    T x = 0;
    T y = 0;
    using R = double;
    R radius = R(t);
    R root = 0.0;
    do {
      x = distribution(random_number_generator);
      y = distribution(random_number_generator);  
      R dx = R(x) * R(x);
      R dy = R(y) * R(y);
      root = std::sqrt(dx + dy);
    } while (root > radius);
    *q = P(x, y);
  }
}

#elif defined(UNIVERSE)

template<typename I>
void generate(I p, I r) {
  std::size_t n = r - p;
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  T t = std::sqrt(n);
  std::uniform_int_distribution<T> distribution(-t, +t);
  for (I q = p; q != r; ++q) {
    T x = distribution(random_number_generator);
    T y = distribution(random_number_generator);
    *q = P(x, y);
  }
}

#elif defined(SPECIAL)

template<typename I>
void generate(I p, I r) {
  assert(std::distance(p, r) >= 4);
  using P = typename std::iterator_traits<I>::value_type;  
  P origin(0, 0);
  P west(-1, 0);
  P north(0, 1);
  P east(1, 0);
  P south(0, -1);
  *p = west;
  ++p;
  *p = north;
  ++p;
  *p = east;
  ++p;
  *p = south;
  ++p;
  for (I q = p; q != r; ++q) {
    *q = origin;
  }
}

#elif defined(LINE)

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  T t = std::numeric_limits<T>::max();
  std::uniform_int_distribution<T> distribution(-t, t);
  for (I q = p; q != r; ++q) {
    T x = distribution(random_number_generator);
    *q = P(x, x);
  }
}

#elif defined(PARABOLA)

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  T j = 0;
  T modulus = 1 << 15;
  for (I q = p; q != r; ++q) {
    *q = P(j, j * j);
    j = (j + 1) % modulus;
  }
  std::shuffle(p, r, random_number_generator);
}

#elif defined(SORTED)

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = int;
  T min = std::numeric_limits<T>::min();
  T max = std::numeric_limits<T>::max();
  std::uniform_int_distribution<T> distribution(min, max);
  for (I q = p; q != r; ++q) {
    T x = distribution(random_number_generator);
    T y = distribution(random_number_generator);
    *q = P(x, y);
  }
  std::sort(p, r,
    [](P const& p, P const& q) -> bool {
      return p.x < q.x;
    });
}

#elif defined(BELL)

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  using D = typename std::iterator_traits<I>::difference_type;
  D n = std::distance(p, r);
  using R = double;
  R radius = R(std::numeric_limits<T>::max());
  R mean = 0.0;
  R stddev = radius / (2 + std::log(n));
  std::normal_distribution<R> distribution(mean, stddev);
  for (I q = p; q != r; ++q) {
    T x = std::round(distribution(random_number_generator));
    T y = std::round(distribution(random_number_generator));
    *q = P(x, y);
  }
}

#else // default SQUARE

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = int; // typename P::coordinate;
  T min = std::numeric_limits<T>::min();
  T max = std::numeric_limits<T>::max();
  std::uniform_int_distribution<T> distribution(min, max);
  for (I q = p; q != r; ++q) {
    T x = distribution(random_number_generator);
    T y = distribution(random_number_generator);
    *q = P(x, y);
  }
}

#endif

void usage(char const* program) {
  std::cerr << "Usage: " << program << " <n>\n";
  exit(1);
}

#if defined(MEASURE_COMPS) or defined(MEASURE_MOVES)

template<typename T>
class counter {
private:

  T datum;

public:

  static constexpr bool is_signed = true;
  static constexpr bool is_integer = true;
  static constexpr bool is_exact = true;
  static constexpr bool is_bounded = true;
  static constexpr bool is_modulo = not is_signed;
  static constexpr std::size_t radix = 2;
  static constexpr std::size_t length = 1;
  static constexpr std::size_t width = 8 * sizeof(T);
  static constexpr T min = std::numeric_limits<T>::min();
  static constexpr T max = std::numeric_limits<T>::max();

  explicit counter() 
    : datum(0) {
    moves += 1;
  }

  counter(int x)
    : datum(x) {
    moves += 1;
  }

  counter(counter const& other) :
    datum(other.datum) {
    moves += 1;
  }

  template<typename number>
  explicit counter(number x = 0) 
    : datum(x) {
    moves += 1;
  }

  counter(counter&& other) {
    datum = std::move(other.datum);
    moves += 1;
  }

  counter& operator=(counter const& other) {
    datum = other.datum;
    moves += 1;
    return *this;
  }

  counter& operator=(counter&& other) {
    datum = std::move(other.datum);
    moves += 1;
    return *this;
  }

  operator T() const {
    return datum;
  }

  template<typename U>
  friend bool operator==(counter<U> const&, counter<U> const&);

  template<typename U>
  friend bool operator!=(counter<U> const&, counter<U> const&);

  template<typename U>
  friend bool operator<(counter<U> const&, counter<U> const&);

  template<typename U>
  friend bool operator>(counter<U> const&, counter<U> const&);

  template<typename U>
  friend bool operator<=(counter<U> const&, counter<U> const&);

  template<typename U>
  friend bool operator>=(counter<U> const&, counter<U> const&);
};

template<typename T>
bool operator==(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum == y.datum;
}

template<typename T>
bool operator!=(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum != y.datum;
}

template<typename T>
bool operator<(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum < y.datum;
}

template<typename T>
bool operator>(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum > y.datum;
}

template<typename T>
bool operator<=(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum <= y.datum;
}

template<typename T>
bool operator>=(counter<T> const& x, counter<T> const& y) {
  ++comps;
  return x.datum >= y.datum;
}

#endif

int main(int argc, char** argv) {
  unsigned long n = 15;
  if (argc == 2) {
    n = std::atoi(argv[1]);
  }
  else {
    usage(argv[0]);
  }

  unsigned long repetitions = maxsize / n;
  if (n > maxsize) {
    repetitions = 1;
    maxsize = n;
  }

#if defined(MEASURE_MOVES) or defined(MEASURE_COMPS)

  using P = point<counter<int>>;

#else

  using P = point<int>;

#endif

  using I = P*;
  I a = new P[maxsize];
  I b = a;
  for (volatile unsigned long t = 0; t != repetitions; ++t) {
    generate(b, b + n);
    b = b + n;
  }

  b = a;

#if defined(MEASURE_TURNS)

  turns = 0;

#elif defined(MEASURE_MOVES)

  moves = 0;

#elif defined(MEASURE_COMPS)

  comps = 0;

#elif defined(PRUNING)

  pruning_efficiency = 0;

#else

  std::clock_t start = std::clock();

#endif

  for (volatile unsigned long t = 0; t != repetitions; ++t)  {
    (void) NAME::solve(&b[0], &b[n]);
    b = b + n;
  }

#if defined(MEASURE_TURNS)

  double t = double(repetitions) * double(n);  
  std::cout.precision(3);
  std::cout << n << "\t" << double(turns) / t << "\n";

#elif defined(MEASURE_MOVES)

  double t = double(repetitions) * double(n);  
  std::cout.precision(3);
  std::cout << n << "\t" << double(moves) / t << "\n";

#elif defined(MEASURE_COMPS)

  double t = double(repetitions) * double(n);  
  std::cout.precision(3);
  std::cout << n << "\t" << double(comps) / t << "\n";

#elif defined(PRUNING)

  double t = double(repetitions) * double(n);
  std::cout.precision(3);
  std::cout << n << "\tpruning: " << double(pruning_efficiency) * 100.0 / t << "\n";

#else

  std::clock_t stop = std::clock();

  double t = double(repetitions) * double(n);
  double ns = 1000000000.0 * double(stop - start) / double(CLOCKS_PER_SEC);
  std::cout.precision(4);
  std::cout << n << "\t" << ns / t << "\n";

#endif

  delete[] a;
  return 0;
}