/*
  Performance Engineering Laboratory © 2017-2018

  Execute the some levels of quickhull and process the remaining
  regions using the plane-sweep algorithm
*/

#ifndef __INTROHULL__
#define __INTROHULL__

#include <algorithm> // std::sort std::partition ...
#include <cassert> // assert macro
#include <cstdlib> // std::size_t
#include <functional> // std::less
#include <iterator> // std::iterator_traits
#include "plane_sweep.h++" // plane_sweep::solve
#include "point.h++" // left_turn
#include <utility> // std::pair std::get ...
#include "validation.h++" // same_multiset convex_polygon all_inside
#include <vector> // std::vector

namespace introhull {

  template<typename N>
  inline N lg(N n) {
    N k;
    for (k = 0; n != 0; n >>= 1) {
      ++k;
    }
    return k - 1;
  }

  inline int lg(int n) {
    using N = std::size_t;
    constexpr N  char_bit = std::numeric_limits<unsigned char>::digits;
    return sizeof(int) * char_bit  - 1 - __builtin_clz(n);
  }

  inline long lg(long n) {
    using N = std::size_t;
    constexpr N char_bit = std::numeric_limits<unsigned char>::digits;
    return sizeof(long) * char_bit - 1 - __builtin_clzl(n);
  }

  inline long long lg(long long n) {
    using N = std::size_t;
    constexpr N char_bit = std::numeric_limits<unsigned char>::digits;
    return sizeof(long long) * char_bit - 1 - __builtin_clzll(n);
  }

  // move *st <- *rd and *nd <- *th by swaps in parallel

  template<typename I>
  void parallel_iter_swap(I st, I nd, I rd, I th) {
    assert(st != nd and rd != th);
    std::iter_swap(st, rd);
    if (th == st) {
      std::iter_swap(nd, rd);
    }
    else {
      std::iter_swap(nd, th);
    }
  }

  template<typename I>
  std::pair<I, I> find_poles(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    auto pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.x < b.x) or (a.x == b.x and a.y < b.y);
      });
    return pair;
  }

  template<typename I>
  I find_furthest(I first, I past, I east) {
    assert(first != past);
    I west = first;
    I answer = west;
    using U = decltype(signed_area(*west, *west, *west));
    U best = 0;
    if ((*east).x == (*west).x) { // vertical
      for (I i = first + 1; i != past; ++i) {
        U _Phi = signed_area(*west, *east, *i);
        if (_Phi > best or (_Phi == best and (*i).y < (*answer).y)) {
          answer = i;
          best = _Phi;
        }
      }
    }
    else {
      for (I i = first + 1; i != past; ++i) {
        U _Phi = signed_area(*west, *east, *i);
        if (_Phi > best or (_Phi == best and (*i).x < (*answer).x)) {
          answer = i;
          best = _Phi;
        }
      }
    }
    return answer;
  }

  template<typename I>
  I partition_left_right(I first, I past, I east) {
    assert(first != past);
    using P = typename std::iterator_traits<I>::value_type;
    I west = first;
    I middle = std::partition(west + 1, past,
      [&](P const& q) -> bool {
        return not left_turn(*west, q, *east);
      });
    return middle;
  }

  template <typename I>
  void swap_blocks(I source, I past_source, I target) {
    using P = typename std::iterator_traits<I>::value_type;
    I hole = target;
    P p = *target;
    I const last = past_source - 1;
    while (true) {
      *hole = *source;
      ++hole;
      if (source == last) {
        break;
      }
      *source = *hole;
      ++source;
    }
    *source = p;
  }

  template <typename I>
  void move_away(I here, I rest, I past) {
    if (here == rest or rest == past) {
      return;
    }
    if (rest - here < past - rest) {
      swap_blocks(here, rest, past - (rest - here));
    }
    else {
      swap_blocks(rest, past, here);
    }
  }

  template<typename I, typename C>
  I recurse(I pole, I past, I antipole, std::size_t depth, C compare) {
    std::size_t n = std::distance(pole, past);
    if (n == 1) {
      return past;
    }
    if (n == 2) {
      if (no_turn(*(pole + 1), *pole, *antipole)) {
        return pole + 1;
      }
      else {
        return past;
      }
    }
    if (depth == 0) {
      return plane_sweep::chain(pole, past, antipole, compare);
    }
    I pivot = find_furthest(pole, past, antipole);
    if (no_turn(*pivot, *pole, *antipole)) {
      return pole + 1;
    }
    I last = past - 1;
    std::iter_swap(pivot, last); // pivot at the end
    I mid = partition_left_right(pole, last, last);
    I eliminated = recurse(pole, mid, last, depth - 1, compare);
    std::iter_swap(mid, last);
    std::iter_swap(eliminated, mid); // pivot at its final place
    pivot = eliminated;
    std::size_t m = past - mid;
    ++mid;
    ++eliminated;
    move_away(eliminated, mid, past);
    I interior = partition_left_right(pivot, pivot + m, antipole); 
    eliminated = recurse(pivot, interior, antipole, depth - 1, compare);
    return eliminated;
  }

  template<typename I>
  I solve(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    std::size_t n = past - first;
    if (n < 2 or (n == 2 and *first != *(first + 1))) {
      return past;
    }
    std::size_t stop = lg(n);
    std::pair<I, I> pair = find_poles(first, past);
    I west = first;
    I east = past - 1;
    parallel_iter_swap(west, east, std::get<0>(pair), std::get<1>(pair));
    if (*west == *east) {
      return first + 1;
    }
    I middle = partition_left_right(west, east, east);
    std::size_t m = past - middle;
    I eliminated = recurse(first, middle, east, stop, std::less<T>());
    std::iter_swap(middle, east);
    std::iter_swap(eliminated, middle); // east at its final place
    east = eliminated;
    ++middle;
    ++eliminated;
    move_away(eliminated, middle, past);
    I rest = recurse(east, east + m, west, stop, std::greater<T>());
    return rest;
  }

  template<typename I>
  bool check(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using S = std::vector<P>;
    using J = typename S::iterator;
    S data;
    std::size_t n = past - first;
    data.resize(n);
    std::copy(first, past, data.begin());
    J rest = solve(data.begin(), data.end());
    bool ok = validation::same_multiset(data.begin(), data.end(), first, past) and validation::convex_polygon(data.begin(), rest) and validation::all_inside(rest, data.end(), data.begin(), rest);
    return ok;
  }
}

#endif