/*
  Performance Engineering Laboratory © 2017-2018
*/

#ifndef __PLANE_SWEEP__
#define __PLANE_SWEEP__

#include <algorithm> // std::sort std::partition ...
#include <cassert> // assert macro
#include <cstdlib> // std::size_t
#include <functional> // std::less std::greater
#include <iterator> // std::iterator_traits
#include "point.h++" // left_turn right_turn
#include <utility> // std::pair std::get ...
#include "validation.h++" // same_multiset convex_polygon all_inside

namespace plane_sweep {

  // move *st <- *rd and *nd <- *th by swaps in parallel

  template<typename I>
  void parallel_iter_swap(I st, I nd, I rd, I th) {
    assert(st != nd and rd != th);
    std::iter_swap(st, rd);
    if (th == st) {
      std::iter_swap(nd, rd);
    }
    else {
      std::iter_swap(nd, th);
    }
  }

  template<typename I>
  std::pair<I, I> find_poles(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    auto pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.x < b.x) or (a.x == b.x and a.y < b.y);
      });
    return pair;
  }

  template<typename I>
  I partition_upper_lower(I pole, I antipole, I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    T upper_limit = std::max((*pole).y, (*antipole).y);
    T lower_limit = std::min((*pole).y, (*antipole).y);
    I i = std::partition(first, past,
      [&](P const& q) -> bool {
        if (q.y >= upper_limit) {
          return true;
        }
        else if (q.y <= lower_limit) {
          return false;
        }
        return not left_turn(*pole, q, *antipole);
      });
    return i;
  }

  template<typename I, typename C>
  std::pair<I, I> clean(I pole, I past, C compare) {
    assert(pole != past);
    using P = typename std::iterator_traits<I>::value_type;
    I k = pole + 1;
    while (k != past and (*k).x == (*pole).x) {
      ++k;
    }
    if (k == pole + 1) {
      return std::make_pair(pole, pole + 1); // (top, next)
    }
    I top = std::max_element(pole + 1, k,
      [&](P const& p, P const& q) -> bool {
        return compare(p.y, q.y);
      });
    if ((*top).y == (*pole).y) {
      return std::make_pair(pole, k);
    }
    std::iter_swap(pole + 1, top);
    return std::make_pair(pole + 1, k);
  }

  template<typename I>
  I swap_blocks(I source, I past_the_end, I target) {
    if (source == target or source == past_the_end) {
      return past_the_end;
    }
    using P = typename std::iterator_traits<I>::value_type;
    P p = *target;
    I const last = past_the_end - 1;
    while (true) {
      *target = *source;
      ++target;
      if (source == last) {
        break;
      }
      *source = *target;
      ++source;
    }
    *source = p;
    return target;
  }

  template<typename I>  
  I scan(I first, I top, I next, I past) {
    assert(first != past);
    for (I i = next; i != past; ++i) {
      while (top != first and not right_turn(*(top - 1), *top, *i)) {
        --top;
      }
      ++top;
      std::iter_swap(i, top);
    }
    return ++top;
  }

  template<typename I>  
  I scan_one_more(I first, I top, I extra) {
    while (top != first and not right_turn(*(top - 1), *top, *extra)) {
      --top;
    }
    return ++top;
  }

  template<typename I, typename C>
  I chain(I pole, I rest, I antipole, C compare) {
    using P = typename std::iterator_traits<I>::value_type;
    if (std::distance(pole, rest) == 1) {
      return rest;
    }
    std::sort(pole + 1, rest,
      [&](P const& p, P const& q) -> bool {
        return compare(p.x, q.x);
      });
    std::pair<I, I> pair = clean(pole, rest, compare);
    I top = std::get<0>(pair);
    I next = std::get<1>(pair);
    I interior = scan(pole, top, next, rest);
    interior = scan_one_more(pole, interior - 1, antipole);
    return interior;
  }

  template<typename I>
  I solve(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    std::size_t n = past - first;
    if (n < 2) {
      return past;
    }
    std::pair<I, I> pair = find_poles(first, past);
    I west = first;
    I east = past - 1;
    parallel_iter_swap(west, east, std::get<0>(pair), std::get<1>(pair)); 
    if (*east == *west) {
      return first + 1;
    }
    I middle = partition_upper_lower(west, east, first + 1, past - 1);
    I rest_1 = chain(west, middle, east, std::less<T>());
    std::iter_swap(middle, east);
    east = middle;
    I rest_2 = chain(east, past, west, std::greater<T>());
    I interior = swap_blocks(east, rest_2, rest_1);
    return interior;
  }

  template<typename I>
  bool check(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using S = std::vector<P>;
    using J = typename S::iterator;
    S data;
    std::size_t n = past - first;
    data.resize(n);
    std::copy(first, past, data.begin());
    J rest = solve(data.begin(), data.end());
    bool ok = validation::same_multiset(data.begin(), data.end(), first, past) and validation::convex_polygon(data.begin(), rest) and validation::all_inside(rest, data.end(), data.begin(), rest);
    return ok;
  }
}

#endif
