/*
  Performance Engineering Laboratory © 2017-2018

  Find the extrema in four directions and eliminate all points inside
  the convex hull of these points.
*/

#ifndef __POLES_FIRST__
#define __POLES_FIRST__

#include <algorithm> // std::sort std::partition ...
#include <cassert> // assert macro
#include <cstdlib> // std::size_t
#include <iterator> // std::iterator_traits
#include "plane_sweep.h++" // plane_sweep::solve
#include "point.h++" // left_turn
#include <utility> // std::iter_swap std::pair std::get ...
#include "validation.h++" // same_multiset convex_polygon all_inside
#include <vector> // std::vector

namespace poles_first {

  template<typename I>
  std::vector<I> find_poles(I first, I past) {
    assert(first != past);
    using P = typename std::iterator_traits<I>::value_type;
    std::vector<I> position;
    position.resize(4);
    std::pair<I, I> pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.x < b.x) or (a.x == b.x and a.y < b.y);
      });
    enum {west = 0, east = 1, south = 2, north = 3};
    position[west] = std::get<0>(pair);
    position[east] = std::get<1>(pair);
    pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.y < b.y) or (a.y == b.y and a.x < b.x);
      });
    position[south] = std::get<0>(pair);
    position[north] = std::get<1>(pair);
    return position;
  }

  template<typename S>
  void remove_duplicates(S& sequence) {
    assert(sequence.size() != 0);
    assert(std::is_sorted(sequence.begin(), sequence.end()));
    std::size_t i = 0;
    std::size_t j = 1;
    auto v = sequence[0];
    while (j != sequence.size()) {
      if (not (sequence[j] == v)) {
        std::swap(sequence[i], v);
        ++i;
        v = sequence[j];
      }
      ++j;
    }
    std::swap(sequence[i], v);
    ++i;
    std::size_t leftovers = sequence.size() - i; 
    while (leftovers != 0) {
      sequence.pop_back();
      --leftovers;
    }
  }

  template<typename P, typename I>
  bool inside_quadrilateral(P const& p, I upper, I lower, I past) {
    assert(past - upper > 2);
    assert(*upper == *(past - 1));
    assert(*upper < *(upper + 1));
    assert(*upper < *(past - 2));
    I i = upper;
    do {
      ++i;
    } while (i != lower and *i < p);
    if (left_turn(*(i - 1), *i, p)) {
      return false;
    }
    I last = past - 1;
    I j = lower;
    do {
      ++j;
    } while (j != last and *j > p);
    if (left_turn(*(j - 1), *j, p)) {
      return false;
    }
    return true;
  }

  template<typename I>
  I remove_inner_part(I first, I past, I polygon, I rest) {
    assert(polygon != rest);
    I i = first;
    if (rest - polygon == 1) {
      for (I j = first; j != past; ++j) {      
        if (not (*j == *polygon)) {
          std::iter_swap(i, j);
          ++i;
        }
      }
      return i;
    }
    if (rest - polygon == 2) {
      for (I j = first; j != past; ++j) {      
        if (not (on_line_segment(*polygon, *(polygon + 1), *j))) {
          std::iter_swap(i, j);
          ++i;
        }
      }
      return i;
    }
    // rest - polygon > 2
    using P = typename std::iterator_traits<I>::value_type;
    std::vector<P> approximate_hull;
    for (I k = polygon; k != rest; ++k) {
      approximate_hull.push_back(*k);
    }
    approximate_hull.push_back(*polygon);
    auto upper = approximate_hull.begin();
    auto lower = upper;
    auto end = approximate_hull.end();
    for (auto k = upper + 1; k != end - 1; ++k) {
      if (*lower < *k) {
        lower = k;
      }
    }
    for (I j = first; j != past; ++j) {
      if (not inside_quadrilateral(*j, upper, lower, end)) {
        std::iter_swap(i, j);
        ++i;
      }
    }
    return i;
  }

  template<typename I>
  I prune(I first, I past) {
    if (past - first < 2) {
      return past;
    }
    std::vector<I> poles = find_poles(first, past);
    std::sort(poles.begin(), poles.end());
    remove_duplicates(poles);
    std::size_t n = poles.size();
    for (std::size_t i = 0; i != n; ++i) {
      std::iter_swap(first + i, poles[i]);
    }
    I rest = plane_sweep::solve(first, first + n);
    I eliminated = remove_inner_part(rest, past, first, rest);
    return eliminated;
  }

  template<typename I>
  I solve(I first, I past) {
    I rest = prune(first, past);
    return plane_sweep::solve(first, rest);
  }

  template<typename I>
  bool check(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using S = std::vector<P>;
    using J = typename S::iterator;
    S data;
    std::size_t n = past - first;
    data.resize(n);
    std::copy(first, past, data.begin());
    J rest = solve(data.begin(), data.end());
    bool ok = validation::same_multiset(data.begin(), data.end(), first, past) and validation::convex_polygon(data.begin(), rest) and validation::all_inside(rest, data.end(), data.begin(), rest);
    return ok;
  }
}

#endif