/*
  Performance Engineering Laboratory © 2017-2018

  Brownie points:
  11 Sep 2017: The test cases for overflow detection by Marius-Florin Cristian
*/

#include <algorithm> // std::copy std::equal
#include <cassert> // assert macro
#include <cstdlib> // std::rand std::size_t
#include <exception> // std::exception
#include <iostream> // std::cout std::cerr
#include <iterator> // std::iterator_traits
#include <limits> // std::numeric_limits
#include <random> // std::linear_congruential_engine
#include <vector> // std::vector

#include "algorithm.h++" // NAME::check
#include "point.h++" // point

using linear_congruential_engine = std::linear_congruential_engine<unsigned long long, 6364136223846793005U, 1442695040888963407U, 0U>;
unsigned long long const seed = 10164167376618180197U;
linear_congruential_engine random_number_generator{seed};

template<typename I>
using checker = bool (*)(I, I);

template<typename A>
using testcase = void (*)(A);

template<typename I>
void generate(I p, I r) {
  using P = typename std::iterator_traits<I>::value_type;
  using T = typename P::coordinate;
  T t = 100;
  std::uniform_int_distribution<T> distribution(-t, t);
  for (I q = p; q != r; ++q) {
    T x = distribution(random_number_generator);
    T y = distribution(random_number_generator);  
    *q = P(x, y);
  }
}

template<typename checker>
void empty_set(checker f) {
  std::cout << "empty set\n";
  using P = point<int>;
  using S = std::vector<P>;
  S input;
  assert(f(input.begin(), input.end()));  
}

template<typename checker>
void one_point(checker f) {
  std::cout << "one point\n";
  using P = point<int>;
  using S = std::vector<P>;
  S input;
  P origo;
  input.push_back(origo);
  assert(f(input.begin(), input.end()));  
}

template<typename checker>
void two_points(checker f) {
  std::cout << "two points\n";
  using P = point<int>;
  using S = std::vector<P>;
  S input;
  P origo;
  P another(1, 1);
  input.push_back(origo);
  input.push_back(another);
  assert(f(input.begin(), input.end()));  
}

template<typename checker>
void three_points_on_a_vertical_line(checker f) {
  std::cout << "three points on a vertical line\n";
  using P = point<int>;
  P origo;
  P st(0, 1);
  P nd(0, 2);
  using S = std::vector<P>;
  S input;
  input.push_back(origo);
  input.push_back(st);
  input.push_back(nd);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void three_points_on_a_horizontal_line(checker f) {
  std::cout << "three points on a horizontal line\n";
  using P = point<int>;
  P origo;
  P up(1, 0);
  P top(2, 0);
  using S = std::vector<P>;
  S input;
  input.push_back(origo);
  input.push_back(up);
  input.push_back(top);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void ten_equal_points(checker f) {
  std::cout << "ten equal points\n";
  using P = point<int>;
  P p(1, 1);
  using S = std::vector<P>;
  S input;
  for (std::size_t i = 0; i != 10; ++i) {
    input.push_back(p);
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void four_poles_and_many_duplicates_inside(checker f) {
  std::cout << "four poles and many duplicates inside\n";
  using P = point<int>;
  using S = std::vector<P>;
  std::size_t n = 10;
  P origin(0, 0);
  P west(-1, 0);
  P north(0, 1);
  P east(1, 0);
  P south(0, -1);
  S input;
  input.push_back(west);
  input.push_back(north);
  input.push_back(east);
  input.push_back(south);
  for (std::size_t i = 0; i != n; ++i) {
    input.push_back(origin);
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void four_poles_with_duplicates(checker f) {
  std::cout << "four poles with duplicates\n";
  using P = point<int>;
  P west(-1, 0);
  P north(0, 1);
  P east(1, 0);
  P south(0, -1);
  using S = std::vector<P>;
  S input;
  for (std::size_t i = 0; i < 5; ++i) {
    input.push_back(south);
    input.push_back(north);
    input.push_back(west);
    input.push_back(east);
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void quadrilateral_with_duplicates(checker f) {
  std::cout << "quadrilateral with duplicates\n";
  using P = point<int>;
  P bottom_left(0, 0);
  P top_left(1, 1);
  P bottom_right(4, 0);
  P top_right(3, 1);
  using S = std::vector<P>;
  S input;
  for (std::size_t i = 0; i < 5; ++i) {
    input.push_back(bottom_left);
    input.push_back(top_left);
    input.push_back(bottom_right);
    input.push_back(top_right);
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void duplicates_on_the_periphery(checker f) {
  std::cout << "duplicates on the periphery\n";
  using P = point<int>;
  using S = std::vector<P>;
  S input;
  for (std::size_t i = 0; i < 2; ++i) {
    input.push_back(P(0, 0));
    input.push_back(P(1, 1));
    input.push_back(P(2, 2));
    input.push_back(P(3, 2));
    input.push_back(P(3, 0));
    input.push_back(P(4, 2));
    input.push_back(P(4, 0));
    input.push_back(P(5, 2));
    input.push_back(P(5, 0));
    input.push_back(P(6, 2));
    input.push_back(P(6, 0));
    input.push_back(P(7, 1));
    input.push_back(P(8, 0));
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void many_east_pole_candidates(checker f) {
  std::cout << "many east-pole candidates\n";
  using P = point<int>;
  P west(0, 0);
  P east(1, 4);
  P three(1, 3);
  P two(1, 2);
  P one(1, 1);
  P zero(1, 0);
  using S = std::vector<P>;
  S input;
  for (std::size_t i = 0; i < 5; ++i) {
    input.push_back(west);
    input.push_back(east);
    input.push_back(three);
    input.push_back(two);
    input.push_back(zero);
    input.push_back(one);
  }
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void line_quadrilateral_line(checker f) {
  std::cout << "line quadrilateral line\n";
  using P = point<int>;
  using S = std::vector<P>;
  S input;
  input.push_back(P(0, 0));
  input.push_back(P(1, 0));
  input.push_back(P(2, 0));
  input.push_back(P(2, 0));
  input.push_back(P(3, 0));

  input.push_back(P(4, 1));
  input.push_back(P(5, 2));
  input.push_back(P(6, 3));
  input.push_back(P(6, 3));
  input.push_back(P(7, 2));
  input.push_back(P(8, 1));

  input.push_back(P(4, -1));
  input.push_back(P(5, -2));
  input.push_back(P(6, -3));
  input.push_back(P(6, -3));
  input.push_back(P(7, -2));
  input.push_back(P(8, -1));

  input.push_back(P(9, 0));
  input.push_back(P(10, 0));
  input.push_back(P(11, 0));
  input.push_back(P(12, 0));
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void many_big_numbers(checker f) {
  std::cout << "many big numbers\n";
  using T = int;
  using P = point<int>;
  T max = std::numeric_limits<T>::max();
  P origo(0, 0);
  P left_bottom(-max, -max);
  P just_above(-max, -max + 1);
  P right_top(max, max);
  P left_top(-max, max);
  P neighbour_below(-max, max - 1);
  P neighbour_beside(-max - 1, max);
  using S = std::vector<P>;
  S input;
  input.push_back(left_bottom);
  input.push_back(just_above);
  input.push_back(origo);
  input.push_back(left_top);
  input.push_back(right_top);
  input.push_back(neighbour_below);
  input.push_back(neighbour_beside);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void noisy_box(checker f) {
  std::cout << "noisy box\n";
  using T = int;
  using P = point<T>;
  T max = std::numeric_limits<T>::max();
  using S = std::vector<P>;
  using I = typename S::iterator;
  constexpr std::size_t n = 10;
  S input(n);
  std::uniform_int_distribution<T> distribution(-max + 1, max - 1);
  for (I q = input.begin(); q != input.end(); ++q) {
    T x = distribution(random_number_generator);
    T y = distribution(random_number_generator);  
    *q = P(x, y);
  }
  input.push_back(P(max, max));
  input.push_back(P(max, -max));
  input.push_back(P(-max, -max));
  input.push_back(P(-max, max));
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void random_points_on_a_parabola(checker f) {
  std::cout << "random points on a parabola\n";
  using P = point<int>;
  using S = std::vector<P>;
  using I = typename S::iterator;
  std::size_t const n = 4;
  S input(n);
  int j = 0;
  for (I q = input.begin(); q != input.end(); ++q) {
    *q = P(j, j * j);
    ++j;
  }
  std::shuffle(input.begin(), input.end(), random_number_generator);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void random_points_in_a_square(checker f) {
  std::cout << "random points in a square\n";
  std::size_t const bign = 10000;
  for (std::size_t n = 0; n <= bign; ++n) {
    if (n % 100 == 0) {
      std::cout << "." << std::flush;
    }
    using P = point<int>;
    using S = std::vector<P>;
    S input(n);
    generate(input.begin(), input.end());
    assert(f(input.begin(), input.end()));
  }
  std::cout << "\n" << std::flush;
}

template<typename checker>
void positive_overflow(checker f) {
  std::cout << "positive overflow\n";
  using T = int;
  using P = point<T>;
  T max = std::numeric_limits<T>::max();
  P origo(100, 100);
  P left_top(100, max);
  P right_top(max, max);
  P middle(100000, 100000);
  P right_bottom(max, 100);
  using S = std::vector<P>;
  S input;
  input.push_back(left_top);
  input.push_back(right_bottom);
  input.push_back(origo);
  input.push_back(middle);
  input.push_back(right_top);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void negative_overflow(checker f) {
  std::cout << "negative overflow\n";
  using T = int;
  using P = point<T>;
  T min = std::numeric_limits<T>::min();
  P origo;
  P left_top(min, 0);
  P left_bottom(min, min);
  P right_bottom(0, min);
  P middle(-10000, -10000);
  using S = std::vector<P>;
  S input;
  input.push_back(left_top);
  input.push_back(right_bottom);
  input.push_back(origo);
  input.push_back(middle);
  input.push_back(left_bottom);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void corners_of_the_universe(checker f) {
  std::cout << "corners of the universe\n";
  using T = int;
  using P = point<T>;
  T min = std::numeric_limits<T>::min();
  T max = std::numeric_limits<T>::max();
  P origo;
  P left_top(min, max);
  P left_bottom(min, min);
  P right_top(max, max);
  P right_bottom(max, min);
  using S = std::vector<P>;
  S input;
  input.push_back(left_top);
  input.push_back(right_bottom);
  input.push_back(origo);
  input.push_back(right_top);
  input.push_back(left_bottom);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void negative_coordinates(checker f) {
  std::cout << "negative coordinates\n";
  using P = point<int>;
  P p1(-3, -3);
  P p2(-3, -2);
  P p3(0, 0);
  P p4(1, -1);
  using S = std::vector<P>;
  S input;
  input.push_back(p1);
  input.push_back(p2);
  input.push_back(p3);
  input.push_back(p4);
  assert(f(input.begin(), input.end()));
}

template<typename checker>
void degenerate_coordinates(checker f) {
  std::cout << "degenerate coordinates\n";
  using P = point<int>;
  P p1(0, 0);
  P p2(1, 0);
  P p3(2, 0);
  P p4(3, 0);
  P p5(3, 1);
  P p6(3, 2);
  P p7(3, 3);
  P p8(2, 3);
  P p9(1, 3);
  P p10(0, 3);
  using S = std::vector<P>;
  S input;
  input.push_back(p1);
  input.push_back(p2);
  input.push_back(p3);
  input.push_back(p4);
  input.push_back(p5);
  input.push_back(p6);
  input.push_back(p7);
  input.push_back(p8);
  input.push_back(p9);
  input.push_back(p10);
  assert(f(input.begin(), input.end()));
}

int main() {
  using P = point<int>;
  using S = std::vector<P>;
  using I = typename S::iterator;
  using R = checker<I>;
  using T = testcase<R>;

  T suite[] = {
    empty_set,
    one_point,
    two_points,
    three_points_on_a_vertical_line,
    three_points_on_a_horizontal_line,
    ten_equal_points,
    four_poles_and_many_duplicates_inside,
    four_poles_with_duplicates,
    quadrilateral_with_duplicates,
    duplicates_on_the_periphery,
    many_east_pole_candidates,
    line_quadrilateral_line,
    many_big_numbers,
    noisy_box,
    random_points_on_a_parabola,
    random_points_in_a_square,
    positive_overflow,
    negative_overflow,
    corners_of_the_universe,
    negative_coordinates,
    degenerate_coordinates
  };

  R program = NAME::check;
  for (T test: suite) {
    try {  
      test(program);
    }  
    catch(std::exception& e) {  
      std::cerr << e.what() << std::endl;  
    }
  }
  return 0;
}