/*
  Performance Engineering Laboratory © 2017-2018

  Find the extrema in eight directions and eliminate all points 
  inside the convex hull of these points.
*/

#ifndef __THROW_AWAY__
#define __THROW_AWAY__

#include <algorithm> // std::sort std::min std::minmax_element
#include <cassert> // assert macro
#include <cmath> // std::sqrt
#include "cphmpl/functions.h++" // cphmpl::width
#include "cphstl/integers.h++" // cphstl::bbbZ
#include <cstddef> // std::size_t
#include <iterator> // std::iterator_traits
#include <limits> // std::numeric_limits
#include "plane_sweep.h++"
#include "point.h++"
#include <utility> // std::pair std::get
#include "validation.h++" // same_multiset convex_polygon all_inside
#include <vector> // std::vector

namespace throw_away {

  template<typename I>
  std::vector<I> find_extrema(I first, I past) {
    assert(first != past);
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
#if defined(MEASURE_MOVES) or defined(MEASURE_COMPS)
#define CONVERSION (int)
#else
#define CONVERSION
#endif
    constexpr std::size_t w = cphmpl::width<T>;
    using Z = cphstl::bbbZ<w + 2>;
    std::vector<I> max_position(8, first);
    T xmin = (*first).x;
    T xmax = (*first).x;
    T ymin = (*first).y;
    T ymax = (*first).y;
    Z ne = Z(CONVERSION xmin) + Z(CONVERSION ymin);
    Z se = Z(CONVERSION xmin) - Z(CONVERSION ymin);
    Z sw = -(Z(CONVERSION xmin) + Z(CONVERSION ymin));
    Z nw = Z(CONVERSION ymin) - Z(CONVERSION xmin);
    for (I i = first; i != past; ++i) {
      T x = (*i).x;
      T y = (*i).y;
      if (x < xmin) {
        xmin = x;
        max_position[0] = i;
      }
      if (x > xmax) {
        xmax = x;
        max_position[1] = i;
      }
      if (y < ymin) {
        ymin = y;
        max_position[2] = i;
      }
      if (y > ymax) {
        ymax = y;
        max_position[3] = i;
      }
#if defined(MEASURE_MOVES)

      ++moves; ++moves; ++moves;

#endif      
      Z dot = Z(CONVERSION x) + Z(CONVERSION y);
#if defined(MEASURE_COMPS)

      ++comps;

#endif      
      if (dot > ne) {
#if defined(MEASURE_MOVES)

      ++moves;

#endif      
        ne = dot;
        max_position[4] = i;
      }
#if defined(MEASURE_MOVES)

      ++moves; ++moves; ++moves;

#endif      
      dot = Z(CONVERSION x) - Z(CONVERSION y);
#if defined(MEASURE_COMPS)

      ++comps;

#endif      
      if (dot > se) {
#if defined(MEASURE_MOVES)

      ++moves;

#endif      
        se = dot;
        max_position[5] = i;
      }
#if defined(MEASURE_MOVES)

      ++moves; ++moves; ++moves;

#endif      
      dot = -(Z(CONVERSION x) + Z(CONVERSION y));
#if defined(MEASURE_COMPS)

      ++comps;

#endif      
      if (dot > sw) {
#if defined(MEASURE_MOVES)

      ++moves;

#endif      
        sw = dot;
        max_position[6] = i;
      }
#if defined(MEASURE_MOVES)

      ++moves; ++moves; ++moves;

#endif      
      dot = Z(CONVERSION y) - Z(CONVERSION x);
#if defined(MEASURE_COMPS)

      ++comps;

#endif      
      if (dot > nw) {
#if defined(MEASURE_MOVES)

      ++moves;

#endif      
        nw = dot;
        max_position[7] = i;
      }
    }
    return max_position;
  }

  template<typename S>
  void remove_duplicates(S& sequence) {
    assert(sequence.size() != 0);
    assert(std::is_sorted(sequence.begin(), sequence.end()));
    std::size_t i = 0;
    std::size_t j = 1;
    auto v = sequence[0];
    while (j != sequence.size()) {
      if (not (sequence[j] == v)) {
        std::swap(sequence[i], v);
        ++i;
        v = sequence[j];
      }
      ++j;
    }
    std::swap(sequence[i], v);
    ++i;
    std::size_t leftovers = sequence.size() - i; 
    while (leftovers != 0) {
      sequence.pop_back();
      --leftovers;
    }
  }

  template<typename P, typename I>
  bool inside_convex_polygon(P const& p, I upper, I lower, I past) {
    assert(past - upper > 2);
    assert(*upper == *(past - 1));
    assert(*upper < *(upper + 1));
    assert(*upper < *(past - 2));
    I i = upper;
    do {
      ++i;
    } while (i != lower and *i < p);
    if (left_turn(*(i - 1), *i, p)) {
      return false;
    }
    I last = past - 1;
    I j = lower;
    do {
      ++j;
    } while (j != last and *j > p);
    if (left_turn(*(j - 1), *j, p)) {
      return false;
    }
    return true;
  }

  template<typename I>
  I eliminate_inner_points(I first, I past, I polygon, I rest) {
    assert(polygon != rest);
    I i = first;
    if (rest - polygon == 1) {
      for (I j = first; j != past; ++j) {      
        if (not (*j == *polygon)) {
          std::iter_swap(i, j);
          ++i;
        }
      }
      return i;
    }
    if (rest - polygon == 2) {
      for (I j = first; j != past; ++j) {      
        if (not (on_line_segment(*polygon, *(polygon + 1), *j))) {
          std::iter_swap(i, j);
          ++i;
        }
      }
      return i;
    }
    // rest - polygon > 2
    using P = typename std::iterator_traits<I>::value_type;
    std::vector<P> approximate_hull;
    for (I k = polygon; k != rest; ++k) {
      approximate_hull.push_back(*k);
    }
    approximate_hull.push_back(*polygon);
    auto upper = approximate_hull.begin();
    auto lower = upper;
    auto end = approximate_hull.end();
    for (auto k = upper + 1; k != end - 1; ++k) {
      if (*lower < *k) {
        lower = k;
      }
    }
    for (I j = first; j != past; ++j) {
      if (not inside_convex_polygon(*j, upper, lower, end)) {
        std::iter_swap(i, j);
        ++i;
      }
    }
    return i;
  }

  template<typename I>
  I prune(I first, I past) {
    if (past - first < 2) {
      return past;
    }
    // Step 1
    std::vector<I> extrema = find_extrema(first, past);
    std::sort(extrema.begin(), extrema.end());
    remove_duplicates(extrema);
    for (std::size_t i = 0; i != extrema.size(); ++i) {
      std::iter_swap(first + i, extrema[i]);
    }
    // Step 2
    I rest = plane_sweep::solve(first, first + extrema.size());
    // Step 3
    I interior = eliminate_inner_points(rest, past, first, rest);
    return interior;
  }

  template<typename I>
  I solve(I first, I past) {
    I rest = prune(first, past);
    return plane_sweep::solve(first, rest);
  }

  template<typename I>
  bool check(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using S = std::vector<P>;
    using J = typename S::iterator;
    S data;
    std::size_t n = past - first;
    data.resize(n);
    std::copy(first, past, data.begin());
    J rest = solve(data.begin(), data.end());
    bool ok = validation::same_multiset(data.begin(), data.end(), first, past) and validation::convex_polygon(data.begin(), rest) and validation::all_inside(rest, data.end(), data.begin(), rest);
    return ok;
  }
}

#endif