/*
  Performance Engineering Laboratory © 2017-2018
*/

#ifndef __TORCH__
#define __TORCH__

#include <algorithm> // std::sort std::partition ...
#include <cassert> // assert macro
#include <cstdlib> // std::size_t
#include <iterator> // std::iterator_traits
#include "point.h++" // right_turn
#include <utility> // std::pair std::get
#include "validation.h++" // same_multiset convex_polygon all_inside

namespace torch {

  // move *st <- *rd and *nd <- *th by swaps in parallel

  template<typename I>
  void parallel_iter_swap(I st, I nd, I rd, I th) {
    assert(st != nd and rd != th);
    std::iter_swap(st, rd);
    if (th == st) {
      std::iter_swap(nd, rd);
    }
    else {
      std::iter_swap(nd, th);
    }
  }

  template<typename I>
  I go_upstairs(I first, I past) {
    if (first == past) {
      return past;
    }
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    T max = (*first).y;
    I i = first + 1;
    for (I j = first + 1; j != past; ++j) {
      if ((*j).y >= max) {
        max = (*j).y;
        std::iter_swap(i, j);
        ++i;
      }
    }
    return i;
  }

  template<typename I>
  I go_downstairs(I first, I past) {
    if (first == past) {
      return past;
    }
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    T min = (*first).y;
    I i = first + 1;
    for (I j = first + 1; j != past; ++j) {
      if ((*j).y <= min) {
        min = (*j).y;
        std::iter_swap(i, j);
        ++i;
      }
    }
    return i;
  }

  template<typename I>  
  I scan(I first, I past) {
    assert(first != past);
    I top = first;
    I next = first + 1;
    for (I i = next; i != past; ++i) {
      while (top != first and not right_turn(*(top - 1), *top, *i)) {
        --top;
      }
      ++top;
      std::iter_swap(i, top);
    }
    return ++top;
  }

  template<typename I>  
  I scan_one_more(I first, I top, I extra) {
    while (top != first and not right_turn(*(top - 1), *top, *extra)) {
      --top;
    }
    return ++top;
  }

  template<typename I>
  I solve(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using T = typename P::coordinate;
    std::size_t n = past - first;
    if (n < 2) {
      return past;
    }

    // find the west and east poles

    std::pair<I, I> pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.x < b.x) or (a.x == b.x and a.y < b.y);
      });
    I west = std::get<0>(pair);
    I east = std::get<1>(pair);

    // all points equal?

    if (*west == *east) {
      return first + 1;
    }

    // all points on a vertical line?

    if ((*west).x == (*east).x) { 
      parallel_iter_swap(first, first + 1, west, east);
      return first + 2;
    }

    // move west to its final place and east temporarily to the end

    I last = past - 1;
    parallel_iter_swap(first, last, west, east);
    west = first;
    east = last;

    // find the south and north poles

    pair = std::minmax_element(first, past,
      [](P const& a, P const& b) -> bool {
        return (a.y < b.y) or (a.y == b.y and a.x < b.x);
      });
    I south = std::get<0>(pair);
    I north = std::get<1>(pair);

    bool east_north_same = (*east == *north);
    bool west_south_same = (*west == *south);
    bool west_north_same = (*west == *north);
    bool east_south_same = (*east == *south);

    // recall the quadrant borders

    T west_y = (*west).y;
    T east_y = (*east).y;
    T north_x = (*north).x;
    T south_x = (*south).x;

    // north-west corner

    I done = west + 1;
    if (not west_north_same) {
      I rest = std::partition(done, last,
        [&](P const& a) -> bool {
          return (a.x <= north_x) and (a.y >= west_y);
        });
      if (rest != done) {
        std::sort(done, rest,
          [](P const& a, P const& b) -> bool {
            return (a.x < b.x);
          });
        rest = go_upstairs(west, rest);
        rest = scan(west, rest);
        done = scan_one_more(west, rest - 1, east);
      }
    }

    // move east to Q_2

    std::iter_swap(done, east); 
    east = done;
    done = done + 1;

    // north-east corner

    if (not east_north_same) {
      I top = east - 1;
      I rest = std::partition(done, past,
        [&](P const& a) -> bool {
          return (a.x >= north_x) and (a.y >= east_y);
        });
      if (rest != done) {
        std::iter_swap(top, east);
        std::swap(top, east);
        std::iter_swap(top, rest - 1);
        top = rest - 1;
        std::sort(east + 1, top,
          [](P const& a, P const& b) -> bool {
            return (a.x > b.x);
          });
        rest = go_upstairs(east, rest);
        std::reverse(east, rest);
        top = east;
        done = scan(top, rest);
        east = done - 1;
      }
    }

    // south-east corner

    if (not east_south_same) {
      I rest = std::partition(done, past,
        [&](P const& a) -> bool {
          return (a.x >= south_x) and (a.y <= east_y);
        });
      if (rest != done) {
        std::sort(done, rest,
          [](P const& a, P const& b) -> bool {
            return (a.x > b.x);
          });
        rest = go_downstairs(east, rest);
        rest = scan(east, rest);
        done = scan_one_more(east, rest - 1, west);
      }
    }

    // south-west corner

    if (not west_south_same) {
      I bottom = done - 1;
      I rest = std::partition(done, past,
        [&](P const& a) -> bool {
          return (a.x <= south_x) and (a.y <= west_y);
        });
      if (rest != done) {
        std::iter_swap(bottom, rest - 1);
        std::sort(bottom, rest - 1,
          [](P const& a, P const& b) -> bool {
            return (a.x < b.x);
          });
        rest = go_downstairs(bottom, rest);
        std::reverse(bottom, rest);
        rest = scan(bottom, rest);
        done = scan_one_more(bottom, rest - 1, west);
      }
    }
    return done;
  }

  template<typename I>
  bool check(I first, I past) {
    using P = typename std::iterator_traits<I>::value_type;
    using S = std::vector<P>;
    using J = typename S::iterator;
    S data;
    std::size_t n = past - first;
    data.resize(n);
    std::copy(first, past, data.begin());
    J rest = solve(data.begin(), data.end());
    bool ok = validation::same_multiset(data.begin(), data.end(), first, past) and validation::convex_polygon(data.begin(), rest) and validation::all_inside(rest, data.end(), data.begin(), rest);
    return ok;
  }
}

#endif